import dataclasses
from typing import TypeVar, Union

import pyarrow as pa
from google.protobuf.message import Message

M = TypeVar("M", bound=Message)

SUPPORTED_ENUM_TYPES = (
    pa.int32(),
    pa.binary(),
    pa.string(),
    pa.dictionary(pa.int32(), pa.string()),
    pa.dictionary(pa.int32(), pa.binary()),
)


@dataclasses.dataclass(frozen=True)
class ProtarrowConfig:
    enum_type: pa.DataType = pa.int32()
    timestamp_type: pa.TimestampType = pa.timestamp("ns", "UTC")
    time_of_day_type: Union[pa.Time64Type, pa.Time32Type] = pa.time64("ns")

    def __post_init__(self):
        assert self.enum_type in SUPPORTED_ENUM_TYPES



import collections.abc
import dataclasses
import datetime
from typing import Any, Callable, Iterable, Iterator, List, Optional, Type, Union

import pyarrow as pa
import pyarrow.compute as pc
from google.protobuf.descriptor import Descriptor, EnumDescriptor, FieldDescriptor
from google.protobuf.internal.containers import MessageMap
from google.protobuf.message import Message
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.wrappers_pb2 import (
    BoolValue,
    BytesValue,
    DoubleValue,
    FloatValue,
    Int32Value,
    Int64Value,
    StringValue,
    UInt32Value,
    UInt64Value,
)
from google.type.date_pb2 import Date
from google.type.timeofday_pb2 import TimeOfDay

from protarrow.common import M

_NANOS_PER_UNIT = {"ns": 1, "us": 1_000, "ms": 1_000_000, "s": 1_000_000_000}
_TIME_CONVERTER = {
    pa.time64("ns"): 1,
    pa.time64("us"): 1_000,
    pa.time32("ms"): 1_000_000,
    pa.time32("s"): 1_000_000_00,
}


def _timestamp_scalar_to_proto(scalar: pa.Scalar) -> Timestamp:
    timestamp = Timestamp()
    value = scalar.value * _NANOS_PER_UNIT[scalar.type.unit]
    timestamp.FromNanoseconds(value)
    return timestamp


def _prepare_array(array: pa.Array) -> pa.Array:
    try:
        # TODO: handle units correctly when this is fixed
        #  https://issues.apache.org/jira/browse/ARROW-18257
        ratio = _TIME_CONVERTER[array.type]
        return pc.multiply(array.cast(pa.int64()), pa.scalar(ratio, pa.int64()))
    except KeyError:
        return array


def _date_scalar_to_proto(scalar: pa.Scalar) -> Date:
    date: datetime.date = scalar.as_py()
    if date == datetime.date.min:
        return Date()
    else:
        return Date(year=date.year, month=date.month, day=date.day)


def _time_of_day_scalar_to_proto(scalar: pa.Scalar) -> TimeOfDay:
    total_nanos = scalar.as_py()
    return TimeOfDay(
        nanos=total_nanos % 1_000_000_000,
        seconds=(total_nanos // 1_000_000_000) % 60,
        minutes=(total_nanos // 60_000_000_000) % 60,
        hours=(total_nanos // 3600_000_000_000),
    )


SPECIAL_TYPES = {
    Timestamp.DESCRIPTOR: _timestamp_scalar_to_proto,
    Date.DESCRIPTOR: _date_scalar_to_proto,
    TimeOfDay.DESCRIPTOR: _time_of_day_scalar_to_proto,
}

NULLABLE_TYPES = (
    BoolValue.DESCRIPTOR,
    BytesValue.DESCRIPTOR,
    DoubleValue.DESCRIPTOR,
    FloatValue.DESCRIPTOR,
    Int32Value.DESCRIPTOR,
    Int64Value.DESCRIPTOR,
    StringValue.DESCRIPTOR,
    UInt32Value.DESCRIPTOR,
    UInt64Value.DESCRIPTOR,
)


def is_custom_field(field_descriptor: FieldDescriptor):
    return (
        field_descriptor.type == FieldDescriptor.TYPE_MESSAGE
        and field_descriptor.message_type not in SPECIAL_TYPES
        and field_descriptor.message_type not in NULLABLE_TYPES
    )


@dataclasses.dataclass(frozen=True)
class OffsetToSize(collections.abc.Iterable):
    array: Union[pa.ListArray, pa.MapArray]

    def __post_init__(self):
        assert pa.types.is_integer(self.array.type)

    def __iter__(self) -> Iterator[int]:
        current_offset = self.array[0].as_py()
        for item in self.array[1:]:
            offset = item.as_py()
            yield offset - current_offset
            current_offset = offset


@dataclasses.dataclass(frozen=True)
class OptionalNestedIterable(collections.abc.Iterable):
    parents: Iterable[Message]
    field_descriptor: FieldDescriptor
    validity_mask: Iterable[pa.BooleanScalar]

    def __iter__(self) -> Iterator[Any]:
        fake_message = self.field_descriptor.message_type._concrete_class()
        for parent, valid in zip(self.parents, self.validity_mask):
            if valid.is_valid and valid.as_py():
                yield getattr(parent, self.field_descriptor.name)
            else:
                yield fake_message


@dataclasses.dataclass(frozen=True)
class RepeatedNestedIterable(collections.abc.Iterable):
    parents: Iterable[Message]
    field_descriptor: FieldDescriptor

    def __post_init__(self):
        assert self.field_descriptor.label == FieldDescriptor.LABEL_REPEATED
        assert self.field_descriptor.type == FieldDescriptor.TYPE_MESSAGE

    def __iter__(self) -> Iterator[Any]:
        for parent in self.parents:
            for child in getattr(parent, self.field_descriptor.name):
                yield child


def convert_scalar(scalar: pa.Scalar) -> Any:
    return scalar.as_py()


def create_enum_converter(
    enum_descriptor: EnumDescriptor, arrow_type: pa.DataType
) -> Callable[[pa.Scalar], int]:
    if pa.types.is_integer(arrow_type):
        return lambda x: x.as_py()
    elif pa.types.is_binary(arrow_type) or (
        pa.types.is_dictionary(arrow_type) and pa.types.is_binary(arrow_type.value_type)
    ):
        mapping = {v.name.encode("utf-8"): v.number for v in enum_descriptor.values}
        return lambda x: mapping.get(x.as_py(), 0)
    elif pa.types.is_string(arrow_type) or (
        pa.types.is_dictionary(arrow_type) and pa.types.is_string(arrow_type.value_type)
    ):
        mapping = {v.name: v.number for v in enum_descriptor.values}
        return lambda x: mapping.get(x.as_py(), 0)
    else:
        raise TypeError(arrow_type)


def get_converter(
    field_descriptor: FieldDescriptor, arrow_type: pa.DataType
) -> Callable[[pa.Scalar], Any]:
    if field_descriptor.type == FieldDescriptor.TYPE_ENUM:
        enum_descriptor: EnumDescriptor = field_descriptor.enum_type
        return create_enum_converter(enum_descriptor, arrow_type)
    elif field_descriptor.type == FieldDescriptor.TYPE_MESSAGE:
        if field_descriptor.message_type in NULLABLE_TYPES:
            return convert_scalar
        else:
            try:
                return SPECIAL_TYPES[field_descriptor.message_type]
            except KeyError:
                raise KeyError(field_descriptor.full_name)
    else:
        return convert_scalar


class PlainAssigner(collections.abc.Iterable):
    def __init__(
        self,
        messages: Iterable[Message],
        field_descriptor: FieldDescriptor,
        arrow_type: pa.DataType,
    ):
        self.messages = messages
        self.field_descriptor = field_descriptor
        self.converter = get_converter(field_descriptor, arrow_type)
        self.nullable = self.field_descriptor.message_type in NULLABLE_TYPES
        self.message = None

    def __iter__(self) -> Iterator[Callable[[pa.Scalar], None]]:
        assert self.message is None
        for message in self.messages:
            self.message = message
            yield self
        self.message = None

    def __call__(self, scalar: pa.Scalar) -> None:
        if scalar.is_valid:
            value = self.converter(scalar)
            if value is not None:
                if self.nullable:
                    getattr(self.message, self.field_descriptor.name).value = value
                else:
                    setattr(self.message, self.field_descriptor.name, value)


class AppendAssigner(collections.abc.Iterable):
    def __init__(
        self,
        messages: Iterable[Message],
        field_descriptor: FieldDescriptor,
        sizes: Iterable[int],
        converter: Callable[[Any], Any],
    ):
        self.messages = messages
        self.field_descriptor = field_descriptor
        assert self.field_descriptor.label == FieldDescriptor.LABEL_REPEATED
        self.sizes = sizes
        self.converter = converter
        self.attribute = None

    def __iter__(self) -> Iterator[Callable[[pa.Scalar], None]]:
        assert self.attribute is None
        for message, size in zip(self.messages, self.sizes):
            self.attribute = getattr(message, self.field_descriptor.name)
            for _ in range(size):
                yield self
        self.attribute = None

    def __call__(self, scalar: pa.Scalar) -> None:
        self.attribute.append(self.converter(scalar))


@dataclasses.dataclass
class MapKeyAssigner(collections.abc.Iterable):
    messages: Iterable[Message]
    field_descriptor: FieldDescriptor
    key_arrow_type: dataclasses.InitVar[pa.DataType]
    offsets: Iterable[int]
    converter: Callable[[pa.Scalar], Any] = dataclasses.field(init=False)
    attribute: Any = None

    def __post_init__(self, key_arrow_type: pa.DataType):
        assert self.field_descriptor.label == FieldDescriptor.LABEL_REPEATED
        assert self.field_descriptor.message_type.GetOptions().map_entry
        self.converter = get_converter(
            self.field_descriptor.message_type.fields_by_name["key"], key_arrow_type
        )

    def __iter__(self) -> Iterator[Callable[[pa.Scalar], Message]]:
        assert self.attribute is None
        for message, offset in zip(self.messages, self.offsets):
            self.attribute = getattr(message, self.field_descriptor.name)
            for _ in range(offset):
                yield self
        self.attribute = None

    def __call__(self, scalar: pa.Scalar) -> Message:
        return self.attribute[self.converter(scalar)]


def _direct_assign_map(attribute: MessageMap, key: Any, value: Any):
    attribute[key] = value


def _merge_assign_map(attribute: MessageMap, key: Any, value: Any):
    if value is None:
        attribute[key]
    else:
        attribute[key].MergeFrom(value)


@dataclasses.dataclass
class MapItemAssigner(collections.abc.Iterable):
    messages: Iterable[Message]
    field_descriptor: FieldDescriptor
    key_arrow_type: dataclasses.InitVar[pa.DataType]
    value_arrow_type: dataclasses.InitVar[pa.DataType]
    offsets: Iterable[int]
    key_converter: Callable[[pa.Scalar], Any] = dataclasses.field(init=False)
    value_converter: Callable[[pa.Scalar], Any] = dataclasses.field(init=False)
    assigner: Callable[[MessageMap, Any, Any], None] = dataclasses.field(init=False)
    attribute: Optional[MessageMap] = None

    def __post_init__(self, key_arrow_type: pa.DataType, value_arrow_type: pa.DataType):
        assert self.field_descriptor.label == FieldDescriptor.LABEL_REPEATED
        assert self.field_descriptor.message_type.GetOptions().map_entry
        self.key_converter = get_converter(
            self.field_descriptor.message_type.fields_by_name["key"], key_arrow_type
        )
        value_descriptor = self.field_descriptor.message_type.fields_by_name["value"]
        self.value_converter = get_converter(value_descriptor, value_arrow_type)
        self.assigner = (
            _merge_assign_map
            if (value_descriptor.type == FieldDescriptor.TYPE_MESSAGE)
            else _direct_assign_map
        )

    def __iter__(self) -> Iterator[Callable[[pa.Scalar, pa.Scalar], Message]]:
        assert self.attribute is None
        for message, offset in zip(self.messages, self.offsets):
            self.attribute = getattr(message, self.field_descriptor.name)
            for _ in range(offset):
                yield self
        self.attribute = None

    def __call__(self, key: pa.Scalar, value: pa.Scalar):
        self.assigner(
            self.attribute,
            self.key_converter(key),
            self.value_converter(value) if value.is_valid else None,
        )


def _extract_struct_field(
    array: pa.StructArray,
    field_descriptor: FieldDescriptor,
    messages: Iterable[Message],
) -> None:
    nested_list = OptionalNestedIterable(messages, field_descriptor, array.is_valid())
    _extract_array_messages(array, field_descriptor.message_type, nested_list)


def _extract_map_field(
    array: pa.MapArray,
    field_descriptor: FieldDescriptor,
    messages: Iterable[Message],
) -> None:
    assert pa.types.is_map(array.type), array.type
    value_descriptor = field_descriptor.message_type.fields_by_name["value"]

    if is_custom_field(value_descriptor):
        # Because protobuf doesn't warranty orders of map,
        # we have to make a copy of the list of values here
        values = []
        for assigner, key in zip(
            MapKeyAssigner(
                messages,
                field_descriptor,
                array.type.key_type,
                OffsetToSize(array.offsets),
            ),
            array.keys,
        ):
            values.append(assigner(key))

        assert pa.types.is_struct(array.type.item_type), array.type
        item_type: pa.StructType = array.type.item_type
        assert isinstance(item_type, pa.StructType)

        for field_descriptor in value_descriptor.message_type.fields:
            field_index = item_type.get_field_index(field_descriptor.name)
            if field_index != -1:
                _extract_field(
                    array.values.field(1).field(field_index),
                    field_descriptor,
                    values,
                )

    else:
        for assigner, key, value in zip(
            MapItemAssigner(
                messages,
                field_descriptor,
                array.type.key_type,
                array.type.item_type,
                OffsetToSize(array.offsets),
            ),
            array.keys,
            _prepare_array(array.values.field(1)),
        ):
            assigner(key, value)


def _extract_repeated_field(
    array: pa.Array,
    field_descriptor: FieldDescriptor,
    messages: Iterable[Message],
) -> None:
    if is_custom_field(field_descriptor):
        if field_descriptor.message_type.GetOptions().map_entry:
            _extract_map_field(array, field_descriptor, messages)
        else:
            _extract_repeated_message(array, field_descriptor, messages)
    else:
        _extract_repeated_primitive(array, field_descriptor, messages)


def _extract_repeated_primitive(
    array: pa.Array, field_descriptor: FieldDescriptor, messages: Iterable[Message]
) -> None:
    assigner = AppendAssigner(
        messages=messages,
        field_descriptor=field_descriptor,
        sizes=OffsetToSize(array.offsets),
        converter=get_converter(field_descriptor, array.type.value_type),
    )

    for each_assigner, value in zip(assigner, _prepare_array(array.values)):
        each_assigner(value)


def _extract_repeated_message(
    array: pa.Array, field_descriptor: FieldDescriptor, messages: Iterable[Message]
):
    assert pa.types.is_list(array.type)
    child = field_descriptor.message_type._concrete_class()
    assigner = AppendAssigner(
        messages,
        field_descriptor,
        OffsetToSize(array.offsets),
        lambda x: x,
    )
    for each_assigner, value in zip(assigner, array.values):
        each_assigner(child)
    _extract_array_messages(
        array.values,
        field_descriptor.message_type,
        RepeatedNestedIterable(messages, field_descriptor),
    )


def _extract_field(
    array: pa.Array, field_descriptor: FieldDescriptor, messages: Iterable[Message]
) -> None:
    array = _prepare_array(array)
    if field_descriptor.label == FieldDescriptor.LABEL_REPEATED:
        _extract_repeated_field(array, field_descriptor, messages)
    elif field_descriptor.message_type in SPECIAL_TYPES:
        extractor = SPECIAL_TYPES[field_descriptor.message_type]
        for message, value in zip(messages, array):
            if value.is_valid:
                getattr(
                    message,
                    field_descriptor.name,
                ).MergeFrom(extractor(value))
    elif (
        field_descriptor.type == FieldDescriptor.TYPE_MESSAGE
        and field_descriptor.message_type not in NULLABLE_TYPES
    ):
        _extract_struct_field(array, field_descriptor, messages)
    else:
        plain_assigner = PlainAssigner(messages, field_descriptor, array.type)
        for plain_assigner, value in zip(plain_assigner, array):
            if value.is_valid:
                plain_assigner(value)


def _extract_record_batch_messages(
    record_batch: pa.RecordBatch,
    message_descriptor: Descriptor,
    messages: Iterable[Message],
) -> None:
    for field_descriptor in message_descriptor.fields:
        if field_descriptor.name in record_batch.schema.names:
            _extract_field(
                record_batch[field_descriptor.name], field_descriptor, messages
            )


def _extract_array_messages(
    array: pa.StructArray,
    message_descriptor: Descriptor,
    messages: Iterable[Message],
) -> None:
    assert pa.types.is_struct(array.type), array.type
    assert isinstance(array, pa.StructArray)
    struct_type: pa.StructType = array.type
    for field_descriptor in message_descriptor.fields:
        index = struct_type.get_field_index(field_descriptor.name)
        if index != -1:
            _extract_field(array.field(index), field_descriptor, messages)


def record_batch_to_messages(
    record_batch: pa.RecordBatch, message_type: Type[M]
) -> List[M]:
    messages = [message_type() for _ in range(record_batch.num_rows)]
    _extract_record_batch_messages(record_batch, message_type.DESCRIPTOR, messages)
    return messages


def table_to_messages(table: pa.Table, message_type: Type[M]) -> List[M]:
    messages = []
    for batch in table.to_reader():
        messages.extend(record_batch_to_messages(batch, message_type))
    return messages



from protarrow.arrow_to_proto import record_batch_to_messages, table_to_messages
from protarrow.common import ProtarrowConfig
from protarrow.proto_to_arrow import (
    message_type_to_schema,
    messages_to_record_batch,
    messages_to_table,
)

__version__ = "0.0.3rc2"
__all__ = [
    "ProtarrowConfig",
    "message_type_to_schema",
    "messages_to_record_batch",
    "messages_to_table",
    "record_batch_to_messages",
    "table_to_messages",
]



"""
Utilities for working with pyarrow.
https://arrow.apache.org/docs/python/
"""
import collections.abc
import dataclasses
import datetime
import operator
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

import pyarrow as pa
import pyarrow.compute as pc
from google.protobuf.descriptor import Descriptor, EnumDescriptor, FieldDescriptor
from google.protobuf.descriptor_pb2 import FieldDescriptorProto
from google.protobuf.internal.containers import MessageMap, RepeatedScalarFieldContainer
from google.protobuf.message import Message
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.wrappers_pb2 import (
    BoolValue,
    BytesValue,
    DoubleValue,
    FloatValue,
    Int32Value,
    Int64Value,
    StringValue,
    UInt32Value,
    UInt64Value,
)
from google.type.date_pb2 import Date
from google.type.timeofday_pb2 import TimeOfDay

from protarrow.common import M, ProtarrowConfig

_PROTO_DESCRIPTOR_TO_PYARROW = {
    Date.DESCRIPTOR: pa.date32(),
    BoolValue.DESCRIPTOR: pa.bool_(),
    BytesValue.DESCRIPTOR: pa.binary(),
    DoubleValue.DESCRIPTOR: pa.float64(),
    FloatValue.DESCRIPTOR: pa.float32(),
    Int32Value.DESCRIPTOR: pa.int32(),
    Int64Value.DESCRIPTOR: pa.int64(),
    StringValue.DESCRIPTOR: pa.string(),
    UInt32Value.DESCRIPTOR: pa.uint32(),
    UInt64Value.DESCRIPTOR: pa.uint64(),
}

_PROTO_PRIMITIVE_TYPE_TO_PYARROW = {
    FieldDescriptorProto.TYPE_DOUBLE: pa.float64(),
    FieldDescriptorProto.TYPE_FLOAT: pa.float32(),
    FieldDescriptorProto.TYPE_INT64: pa.int64(),
    FieldDescriptorProto.TYPE_UINT64: pa.uint64(),
    FieldDescriptorProto.TYPE_INT32: pa.int32(),
    FieldDescriptorProto.TYPE_FIXED64: pa.uint64(),
    FieldDescriptorProto.TYPE_FIXED32: pa.uint32(),
    FieldDescriptorProto.TYPE_BOOL: pa.bool_(),
    FieldDescriptorProto.TYPE_STRING: pa.string(),
    FieldDescriptorProto.TYPE_BYTES: pa.binary(),
    FieldDescriptorProto.TYPE_UINT32: pa.uint32(),
    FieldDescriptorProto.TYPE_ENUM: pa.binary(),
    FieldDescriptorProto.TYPE_SFIXED32: pa.int32(),
    FieldDescriptorProto.TYPE_SFIXED64: pa.int64(),
    FieldDescriptorProto.TYPE_SINT32: pa.int32(),
    FieldDescriptorProto.TYPE_SINT64: pa.int64(),
}


def _time_of_day_to_nanos(time_of_day: TimeOfDay) -> int:
    return (
        (time_of_day.hours * 60 + time_of_day.minutes) * 60 + time_of_day.seconds
    ) * 1_000_000_000 + time_of_day.nanos


def _time_of_day_to_micros(time_of_day: TimeOfDay) -> int:
    return (
        (time_of_day.hours * 60 + time_of_day.minutes) * 60 + time_of_day.seconds
    ) * 1_000_000 + time_of_day.nanos // 1_000


def _time_of_day_to_millis(time_of_day: TimeOfDay) -> int:
    return (
        (time_of_day.hours * 60 + time_of_day.minutes) * 60 + time_of_day.seconds
    ) * 1_000 + time_of_day.nanos // 1_000_000


def _time_of_day_to_seconds(time_of_day: TimeOfDay) -> int:
    return (time_of_day.hours * 60 + time_of_day.minutes) * 60 + time_of_day.seconds


def _proto_date_to_py_date(proto_date: Date) -> datetime.date:
    # TODO: fix when
    x: datetime.date.min
    if proto_date.year == 0:
        return datetime.date.min
    else:
        return datetime.date(proto_date.year, proto_date.month, proto_date.day)


_PROTO_DESCRIPTOR_TO_ARROW_CONVERTER = {
    Date.DESCRIPTOR: _proto_date_to_py_date,
    TimeOfDay.DESCRIPTOR: _time_of_day_to_nanos,
    BoolValue.DESCRIPTOR: operator.attrgetter("value"),
    BytesValue.DESCRIPTOR: operator.attrgetter("value"),
    DoubleValue.DESCRIPTOR: operator.attrgetter("value"),
    FloatValue.DESCRIPTOR: operator.attrgetter("value"),
    Int32Value.DESCRIPTOR: operator.attrgetter("value"),
    Int64Value.DESCRIPTOR: operator.attrgetter("value"),
    StringValue.DESCRIPTOR: operator.attrgetter("value"),
    UInt32Value.DESCRIPTOR: operator.attrgetter("value"),
    UInt64Value.DESCRIPTOR: operator.attrgetter("value"),
}

_TIMESTAMP_CONVERTERS = {
    "s": Timestamp.ToSeconds,
    "ms": Timestamp.ToMilliseconds,
    "us": Timestamp.ToMicroseconds,
    "ns": Timestamp.ToNanoseconds,
}

_TIME_OF_DAY_CONVERTERS = {
    "s": _time_of_day_to_seconds,
    "ms": _time_of_day_to_millis,
    "us": _time_of_day_to_micros,
    "ns": _time_of_day_to_nanos,
}


@dataclasses.dataclass(frozen=True)
class FlattenedIterable(collections.abc.Iterable):
    parents: Iterable[Iterable[Optional[Any]]]

    def __iter__(self) -> Iterator[Any]:
        for parent in self.parents:
            if parent is not None:
                for child in parent:
                    yield child


@dataclasses.dataclass(frozen=True)
class NestedIterable(collections.abc.Iterable):
    parents: Iterable[Optional[Message]]
    getter: Callable[[Message], Any]

    def __iter__(self) -> Iterator[Optional[Any]]:
        for parent in self.parents:
            if parent is not None:
                yield self.getter(parent)
            else:
                yield None


@dataclasses.dataclass(frozen=True)
class NestedMessageGetter:
    name: str

    def __call__(self, message: Message) -> Optional[Message]:
        if message.HasField(self.name):
            return getattr(message, self.name)
        else:
            return None


@dataclasses.dataclass(frozen=True)
class MapKeyIterable(collections.abc.Iterable):
    scalar_map: Iterable[Optional[MessageMap]]

    def __iter__(self) -> Iterator[Any]:
        for scalar_map in self.scalar_map:
            if scalar_map is not None:
                for key in scalar_map.keys():
                    yield key


@dataclasses.dataclass(frozen=True)
class MapValueIterable(collections.abc.Iterable):
    scalar_map: Iterable[Optional[MessageMap]]

    def __iter__(self) -> Iterator[Any]:
        for scalar_map in self.scalar_map:
            if scalar_map is not None:
                for value in scalar_map.values():
                    yield value


def get_enum_converter(
    data_type: pa.DataType, enum_descriptor: EnumDescriptor
) -> Callable[[int], Any]:
    if data_type == pa.int32():
        return lambda x: x

    elif data_type == pa.binary() or data_type == pa.dictionary(
        pa.int32(), pa.binary()
    ):
        values_by_number = {
            k: bytes(v.name, encoding="UTF-8")
            for k, v in enum_descriptor.values_by_number.items()
        }

        def converter(x: int) -> bytes:
            return values_by_number[x]

        return converter
    elif data_type == pa.string() or data_type == pa.dictionary(
        pa.int32(), pa.string()
    ):
        values_by_number = {
            k: v.name for k, v in enum_descriptor.values_by_number.items()
        }

        def converter(x: int) -> bytes:
            return values_by_number[x]

        return converter
    else:
        raise TypeError(data_type)


def _get_converter_and_type(
    field_descriptor: FieldDescriptor,
    config: ProtarrowConfig,
) -> Optional[Tuple[pa.DataType, Callable[[Any], Any]]]:
    if field_descriptor.message_type == Timestamp.DESCRIPTOR:
        return config.timestamp_type, _TIMESTAMP_CONVERTERS[config.timestamp_type.unit]
    elif field_descriptor.message_type == TimeOfDay.DESCRIPTOR:
        return (
            config.time_of_day_type,
            _TIME_OF_DAY_CONVERTERS[config.time_of_day_type.unit],
        )
    elif field_descriptor.type == FieldDescriptorProto.TYPE_MESSAGE:
        converter = _PROTO_DESCRIPTOR_TO_ARROW_CONVERTER.get(
            field_descriptor.message_type
        )
        if converter is None:
            return None
        else:
            return (
                _PROTO_DESCRIPTOR_TO_PYARROW[field_descriptor.message_type],
                converter,
            )
    elif field_descriptor.type == FieldDescriptorProto.TYPE_ENUM:
        return config.enum_type, get_enum_converter(
            config.enum_type, field_descriptor.enum_type
        )
    elif field_descriptor.type in _PROTO_PRIMITIVE_TYPE_TO_PYARROW:

        def converter(x: Any) -> Any:
            return x

        return _PROTO_PRIMITIVE_TYPE_TO_PYARROW.get(field_descriptor.type), converter
    else:
        raise TypeError(
            f"Unsupported field type "
            f"{FieldDescriptorProto.Type.Name(field_descriptor.type)} "
            f"for {field_descriptor.name}"
        )


def _proto_field_to_array(
    records: Iterable[Message],
    field: FieldDescriptor,
    validity_mask: Optional[Iterable[bool]],
    config: ProtarrowConfig,
) -> pa.Array:
    type_and_converter = _get_converter_and_type(field, config)
    if type_and_converter is not None:
        pa_type, converter = type_and_converter

        null_value = (
            None
            if (
                field.type == FieldDescriptor.TYPE_MESSAGE
                # We use none for repeated field as there should not
                # be any missing list elements, they are not nullable
                or field.label == FieldDescriptor.LABEL_REPEATED
            )
            else converter(field.default_value)
        )
        array = []
        for i, record in enumerate(records):
            if record is None or not (validity_mask is None or validity_mask[i]):
                value = null_value
            else:
                value = converter(record)
            array.append(value)
        return pa.array(array, pa_type)
    else:
        return _message_to_array(
            records,
            field.message_type,
            validity_mask=validity_mask,
            config=config,
        )


def _get_offsets(
    records: Iterable[Union[RepeatedScalarFieldContainer, MessageMap]]
) -> List[int]:
    last_offset = 0
    offsets = []
    for record in records:
        if record is None:
            offsets.append(None)
        else:
            offsets.append(last_offset)
            last_offset += len(record)
    offsets.append(last_offset)
    return offsets


def _repeated_proto_to_array(
    records: Iterable[RepeatedScalarFieldContainer],
    field: FieldDescriptor,
    config: ProtarrowConfig,
) -> pa.ListArray:
    """
    Convert Protobuf embedded lists to a 1-dimensional PyArrow ListArray with offsets
    See PyArrow Layout format documentation on how to calculate offsets.
    """
    offsets = _get_offsets(records)
    array = _proto_field_to_array(FlattenedIterable(records), field, None, config)
    return pa.ListArray.from_arrays(
        offsets,
        array,
        pa.list_(pa.field("item", array.type, nullable=False)),
    )


def _proto_map_to_array(
    records: Iterable[MessageMap],
    field: FieldDescriptor,
    config: ProtarrowConfig = ProtarrowConfig(),
) -> pa.MapArray:
    """
    Convert Protobuf maps to a 1-dimensional PyArrow MapArray with offsets
    See PyArrow Layout format documentation on how to calculate offsets.
    """
    key_field = field.message_type.fields_by_name["key"]
    value_field = field.message_type.fields_by_name["value"]
    offsets = _get_offsets(records)
    keys = _proto_field_to_array(
        MapKeyIterable(records),
        key_field,
        validity_mask=None,
        config=config,
    )
    values = _proto_field_to_array(
        MapValueIterable(records),
        value_field,
        validity_mask=None,
        config=config,
    )
    return pa.MapArray.from_arrays(offsets, keys, values).cast(
        pa.map_(keys.type, pa.field("item", values.type, nullable=False))
    )


def _proto_field_nullable(field: FieldDescriptor) -> bool:
    return (
        field.type == FieldDescriptorProto.TYPE_MESSAGE
        and field.label != FieldDescriptorProto.LABEL_REPEATED
    )


def _proto_field_validity_mask(
    records: Iterable[Message], field: FieldDescriptor
) -> Optional[List[bool]]:
    if (
        field.type != FieldDescriptorProto.TYPE_MESSAGE
        or field.label == FieldDescriptorProto.LABEL_REPEATED
    ):
        return None
    mask = []
    field_name = field.name
    for record in records:
        if record is None:
            mask.append(False)
        else:
            mask.append(record.HasField(field_name))
    return mask


def _message_to_array(
    records: Iterable[Message],
    descriptor: Descriptor,
    validity_mask: Optional[Sequence[bool]],
    config: ProtarrowConfig,
) -> pa.StructArray:
    arrays = []
    fields = []

    for field in descriptor.fields:
        if (
            field.type == FieldDescriptor.TYPE_MESSAGE
            and field.label != FieldDescriptor.LABEL_REPEATED
        ):
            field_values = NestedIterable(records, NestedMessageGetter(field.name))
        else:
            field_values = NestedIterable(records, operator.attrgetter(field.name))
        if field.message_type and field.message_type.GetOptions().map_entry:
            array = _proto_map_to_array(field_values, field, config)
        elif field.label == FieldDescriptorProto.LABEL_REPEATED:
            array = _repeated_proto_to_array(field_values, field, config)
        else:
            mask = _proto_field_validity_mask(records, field)
            array = _proto_field_to_array(
                field_values, field, validity_mask=mask, config=config
            )

        arrays.append(array)
        fields.append(
            pa.field(
                field.name,
                array.type,
                nullable=_proto_field_nullable(field),
            )
        )
    return pa.StructArray.from_arrays(
        arrays=arrays,
        fields=fields,
        mask=pc.invert(pa.array(validity_mask, pa.bool_())) if validity_mask else None,
    )


def messages_to_record_batch(
    records: Iterable[M],
    message_type: Type[M],
    config: ProtarrowConfig = ProtarrowConfig(),
):
    return pa.RecordBatch.from_struct_array(
        _message_to_array(
            records,
            message_type.DESCRIPTOR,
            validity_mask=None,
            config=config,
        )
    )


def messages_to_table(
    records: Iterable[M],
    message_type: Type[M],
    config: ProtarrowConfig = ProtarrowConfig(),
) -> pa.Table:
    """Converts a list of protobuf messages to a `pa.Table`"""
    assert isinstance(config, ProtarrowConfig), config
    record_batch = messages_to_record_batch(records, message_type, config=config)
    return pa.Table.from_batches([record_batch])


def message_type_to_schema(
    message_type: Type[M], config: ProtarrowConfig = ProtarrowConfig()
) -> pa.Schema:
    return messages_to_record_batch([], message_type, config).schema



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['protarrow']

package_data = \
{'': ['*']}

install_requires = \
['googleapis-common-protos>=1.53.0,<2.0.0',
 'protobuf>=3.20.1,<4.0.0',
 'pyarrow>=8.0.0']

setup_kwargs = {
    'name': 'protarrow',
    'version': '0.0.3rc2',
    'description': 'Convert from protobuf to arrow and back',
    'long_description': '[![PyPI Version][pypi-image]][pypi-url]\n[![][versions-image]][versions-url]\n[![][stars-image]][stars-url]\n[![codecov](https://codecov.io/gh/0x26res/protarrow/branch/master/graph/badge.svg?token=XMFH27IL70)](https://codecov.io/gh/0x26res/protarrow)\n[![Build Status][build-image]][build-url]\n[![Documentation][doc-image]][doc-url]\n\n\n# Protarrow\n\nA python library for converting from protobuf to arrow and back \n\n# Installation\n\n```shell\npip install protarrow\n```\n\n# Usage\n\nSee the [documentation](https://protarrow.readthedocs.io/en/latest/)\n\n\n<!-- Badges: -->\n\n[pypi-image]: https://img.shields.io/pypi/v/protarrow\n[pypi-url]: https://pypi.org/project/protarrow/\n[build-image]: https://github.com/tradewelltech/protarrow/actions/workflows/build.yaml/badge.svg\n[build-url]: https://github.com/tradewelltech/protarrow/actions/workflows/build.yaml\n[stars-image]: https://img.shields.io/github/stars/tradewelltech/protarrow\n[stars-url]: https://github.com/tradewelltech/protarrow\n[versions-image]: https://img.shields.io/pypi/pyversions/protarrow\n[versions-url]: https://pypi.org/project/protarrow/\n[doc-image]: https://readthedocs.org/projects/protarrow/badge/?version=latest\n[doc-url]: https://protarrow.readthedocs.io/en/latest/?badge=latest\n',
    'author': '0x26res',
    'author_email': '0x26res@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/tradewelltech/protarrow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
